from dotenv import load_dotenv
load_dotenv()  # take environment variables from .env.

import asyncio
import threading
import os
import pyaudio
from starlette.websockets import WebSocket
from queue import Queue
from pynput import keyboard
import json
import traceback
import websockets
import queue
import pydub
import ast
from pydub import AudioSegment
from pydub.playback import play
import io
import time
import wave
import tempfile
from datetime import datetime
from interpreter import interpreter # Just for code execution. Maybe we should let people do from interpreter.computer import run?
from ..server.utils.kernel import put_kernel_messages_into_queue
from ..server.utils.get_system_info import get_system_info
from ..server.stt.stt import stt_wav

from ..server.utils.logs import setup_logging
from ..server.utils.logs import logger
setup_logging()

# Configuration for Audio Recording
CHUNK = 1024  # Record in chunks of 1024 samples
FORMAT = pyaudio.paInt16  # 16 bits per sample
CHANNELS = 1  # Mono
RATE = 44100  # Sample rate
RECORDING = False  # Flag to control recording state
SPACEBAR_PRESSED = False  # Flag to track spacebar press state

# Specify OS
current_platform = get_system_info()

# Initialize PyAudio
p = pyaudio.PyAudio()

import asyncio

send_queue = queue.Queue()

class Device:
    def __init__(self):
        pass

    def record_audio(self):
        
        if os.getenv('STT_RUNNER') == "server":
            # STT will happen on the server. we're sending audio.
            send_queue.put({"role": "user", "type": "audio", "format": "audio/wav", "start": True})
        elif os.getenv('STT_RUNNER') == "client":
            # STT will happen here, on the client. we're sending text.
            send_queue.put({"role": "user", "type": "message", "start": True})
        else:
            raise Exception("STT_RUNNER must be set to either 'client' or 'server'.")

        """Record audio from the microphone and add it to the queue."""
        stream = p.open(format=FORMAT, channels=CHANNELS, rate=RATE, input=True, frames_per_buffer=CHUNK)
        logger.info("Recording started...")
        global RECORDING

        # Create a temporary WAV file to store the audio data
        temp_dir = tempfile.gettempdir()
        wav_path = os.path.join(temp_dir, f"audio_{datetime.now().strftime('%Y%m%d%H%M%S%f')}.wav")
        wav_file = wave.open(wav_path, 'wb')
        wav_file.setnchannels(CHANNELS)
        wav_file.setsampwidth(p.get_sample_size(FORMAT))
        wav_file.setframerate(RATE)

        while RECORDING:
            data = stream.read(CHUNK, exception_on_overflow=False)
            wav_file.writeframes(data)

        wav_file.close()
        stream.stop_stream()
        stream.close()
        logger.info("Recording stopped.")

        duration = wav_file.getnframes() / RATE
        if duration < 0.3:
            # Just pressed it. Send stop message
            if os.getenv('STT_RUNNER') == "client":
                send_queue.put({"role": "user", "type": "message", "content": "stop"})
                send_queue.put({"role": "user", "type": "message", "end": True})
            else:
                send_queue.put({"role": "user", "type": "audio", "format": "audio/wav", "content": ""})
                send_queue.put({"role": "user", "type": "audio", "format": "audio/wav", "end": True})
        else:
            if os.getenv('STT_RUNNER') == "client":
                # Run stt then send text
                text = stt_wav(wav_path)
                send_queue.put({"role": "user", "type": "message", "content": text})
                send_queue.put({"role": "user", "type": "message", "end": True})
            else:
                # Stream audio
                with open(wav_path, 'rb') as audio_file:
                    byte_data = audio_file.read(CHUNK)
                    while byte_data:
                        send_queue.put({"role": "user", "type": "audio", "format": "audio/wav", "content": str(byte_data)})
                        byte_data = audio_file.read(CHUNK)
                send_queue.put({"role": "user", "type": "audio", "format": "audio/wav", "end": True})

        if os.path.exists(wav_path):
            os.remove(wav_path)

    def toggle_recording(self, state):
        """Toggle the recording state."""
        global RECORDING, SPACEBAR_PRESSED
        if state and not SPACEBAR_PRESSED:
            SPACEBAR_PRESSED = True
            if not RECORDING:
                RECORDING = True
                threading.Thread(target=self.record_audio).start()
        elif not state and SPACEBAR_PRESSED:
            SPACEBAR_PRESSED = False
            RECORDING = False

    def on_press(self, key):
        """Detect spacebar press."""
        if key == keyboard.Key.space:
            self.toggle_recording(True)

    def on_release(self, key):
        """Detect spacebar release and ESC key press."""
        if key == keyboard.Key.space:
            self.toggle_recording(False)
        elif key == keyboard.Key.esc:
            logger.info("Exiting...")
            os._exit(0)

    async def message_sender(self, websocket):
        while True:
            message = await asyncio.get_event_loop().run_in_executor(None, send_queue.get)
            await websocket.send(json.dumps(message))
            send_queue.task_done()

    async def websocket_communication(self, WS_URL):
        while True:
            try:
                async with websockets.connect(WS_URL) as websocket:
                    logger.info("Press the spacebar to start/stop recording. Press ESC to exit.")
                    asyncio.create_task(self.message_sender(websocket))

                    initial_message = {"role": None, "type": None, "format": None, "content": None} 
                    message_so_far = initial_message

                    while True:
                        message = await websocket.recv()

                        logger.debug(f"Got this message from the server: {type(message)} {message}")

                        if type(message) == str:
                            message = json.loads(message)

                        if message.get("end"):
                            logger.debug(f"Complete message from the server: {message_so_far}")
                            logger.info("\n")
                            message_so_far = initial_message

                        if "content" in message:
                            print(message['content'], end="", flush=True)
                            if any(message_so_far[key] != message[key] for key in message_so_far if key != "content"):
                                message_so_far = message
                            else:
                                message_so_far["content"] += message["content"]

                        if message["type"] == "audio" and "content" in message:
                            audio_bytes = bytes(ast.literal_eval(message["content"]))

                            # Convert bytes to audio file
                            audio_file = io.BytesIO(audio_bytes)
                            audio = AudioSegment.from_mp3(audio_file)

                            # Play the audio
                            play(audio)

                            await asyncio.sleep(1)

                        # Run the code if that's the client's job
                        if os.getenv('CODE_RUNNER') == "client":
                            if message["type"] == "code" and "end" in message:
                                language = message_so_far["format"]
                                code = message_so_far["content"]
                                result = interpreter.computer.run(language, code)
                                send_queue.put(result)
    

            except:
                # traceback.print_exc()
                logger.info(f"Connecting to `{WS_URL}`...")
                await asyncio.sleep(2)

    async def start_async(self):
            # Configuration for WebSocket
            WS_URL = os.getenv('SERVER_URL')
            if not WS_URL:
                raise ValueError("The environment variable SERVER_URL is not set. Please set it to proceed.")

            # Start the WebSocket communication
            asyncio.create_task(self.websocket_communication(WS_URL))

            # Start watching the kernel if it's your job to do that
            if os.getenv('CODE_RUNNER') == "client":
                asyncio.create_task(put_kernel_messages_into_queue(send_queue))

            
            # If Raspberry Pi, add the button listener, otherwise use the spacebar
            if current_platform.startswith("raspberry-pi"):
                logger.info("Raspberry Pi detected, using button on GPIO pin 15")
                # Use GPIO pin 15
                pindef = ["gpiochip4", "15"] # gpiofind PIN15
                print("PINDEF", pindef)

                # HACK: needs passwordless sudo
                process = await asyncio.create_subprocess_exec("sudo", "gpiomon", "-brf", *pindef, stdout=asyncio.subprocess.PIPE)
                while True:
                    line = await process.stdout.readline()
                    if line:
                        line = line.decode().strip()
                        if "FALLING" in line:
                            self.toggle_recording(False)
                        elif "RISING" in line:
                            self.toggle_recording(True)
                    else:
                        break
            else:
                # Keyboard listener for spacebar press/release
                listener = keyboard.Listener(on_press=self.on_press, on_release=self.on_release)
                listener.start()

    def start(self):
        asyncio.run(self.start_async())
        p.terminate()


"""
Defines a function which takes text and returns a path to an audio file.
"""

from pydub import AudioSegment
from dotenv import load_dotenv
load_dotenv()  # take environment variables from .env.
import tempfile
from openai import OpenAI
import os
import subprocess
import tempfile
from pydub import AudioSegment
from pydub.playback import play
import simpleaudio as sa

client = OpenAI()

def tts(text, play_audio):
    if os.getenv('ALL_LOCAL') == 'False':
        response = client.audio.speech.create(
            model="tts-1",
            voice="alloy",
            input=text,
            response_format="mp3"
        )
        with tempfile.NamedTemporaryFile(suffix=".mp3") as temp_file:
            response.stream_to_file(temp_file.name)
            
            if play_audio:
                audio = AudioSegment.from_mp3(temp_file.name)
                play(audio)

            
            return temp_file.read()
    else:
        with tempfile.NamedTemporaryFile(suffix=".wav", delete=False) as temp_file:
            output_file = temp_file.name
            piper_dir = os.path.join(os.path.dirname(__file__), 'local_tts', 'piper')
            subprocess.run([
                os.path.join(piper_dir, 'piper'),
                '--model', os.path.join(piper_dir, os.getenv('PIPER_VOICE_NAME')),
                '--output_file', output_file
            ], input=text, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)

            if play_audio:
                audio = AudioSegment.from_mp3(temp_file.name)
                # Apply a 500ms fade out
                audio = audio.fade_out(500)

                # Add 500ms of silence to the end
                silence = Silence().to_audio_segment(duration=500)
                audio_with_padding = audio + silence

                play(audio_with_padding)
            return temp_file.read()

def play_mp3(filename):

    # Load your MP3 audio file
    audio = AudioSegment.from_mp3("your_audio_file.mp3")

    # Apply a fade-out (optional but recommended to smooth the end)
    audio = audio.fade_out(500)

    # Add silence at the end
    silence_duration_ms = 500  # Duration of silence in milliseconds
    silence = AudioSegment.silent(duration=silence_duration_ms)
    audio_with_padding = audio + silence

    # Save the modified audio as a WAV file for compatibility with simpleaudio
    audio_with_padding.export("output_audio.wav", format="wav")

    # Load the processed WAV file
    wave_obj = sa.WaveObject.from_wave_file("output_audio.wav")

    # Play the audio
    play_obj = wave_obj.play()

    # Wait for the playback to finish
    play_obj.wait_done()




from dotenv import load_dotenv
load_dotenv()  # take environment variables from .env.

import ast
import json
import queue
import os
import traceback
import re
from fastapi import FastAPI
from fastapi.responses import PlainTextResponse
from starlette.websockets import WebSocket
from .stt.stt import stt_bytes
from .tts.tts import tts
from pathlib import Path
import asyncio
import urllib.parse
from .utils.kernel import put_kernel_messages_into_queue
from .i import configure_interpreter
from interpreter import interpreter
import ngrok

from .utils.logs import setup_logging
from .utils.logs import logger
setup_logging()


app = FastAPI()

conversation_history_path = Path(__file__).parent / 'conversations' / 'user.json'

# This is so we only say() full sentences
def is_full_sentence(text):
    return text.endswith(('.', '!', '?'))

def split_into_sentences(text):
    return re.split(r'(?<=[.!?])\s+', text)

# Queues
from_computer = queue.Queue() # Just for computer messages from the device. Sync queue because interpreter.run is synchronous
from_user = asyncio.Queue() # Just for user messages from the device.
to_device = asyncio.Queue() # For messages we send.

# Switch code executor to device if that's set

if os.getenv('CODE_RUNNER') == "device":

    # (This should probably just loop through all languages and apply these changes instead)

    class Python:
        # This is the name that will appear to the LLM.
        name = "python"

        def __init__(self):
            self.halt = False

        def run(self, code):
            """Generator that yields a dictionary in LMC Format."""

            # Prepare the data
            message = {"role": "assistant", "type": "code", "format": "python", "content": code}

            # Unless it was just sent to the device, send it wrapped in flags
            if not (interpreter.messages and interpreter.messages[-1] == message):
                to_device.put({"role": "assistant", "type": "code", "format": "python", "start": True})
                to_device.put(message)
                to_device.put({"role": "assistant", "type": "code", "format": "python", "end": True})
            
            # Stream the response
            logger.info("Waiting for the device to respond...")
            while True:
                chunk = from_computer.get()
                logger.info(f"Server received from device: {chunk}")
                if "end" in chunk:
                    break
                yield chunk

        def stop(self):
            self.halt = True

        def terminate(self):
            """Terminates the entire process."""
            # dramatic!! do nothing
            pass

    interpreter.computer.languages = [Python]

# Configure interpreter
interpreter = configure_interpreter(interpreter)

@app.get("/ping")
async def ping():
    return PlainTextResponse("pong")

@app.websocket("/")
async def websocket_endpoint(websocket: WebSocket):
    await websocket.accept()
    receive_task = asyncio.create_task(receive_messages(websocket))
    send_task = asyncio.create_task(send_messages(websocket))
    try:
        await asyncio.gather(receive_task, send_task)
    except Exception as e:
        traceback.print_exc()
        logger.info(f"Connection lost. Error: {e}")

async def receive_messages(websocket: WebSocket):
    while True:
        data = await websocket.receive_json()
        if data["role"] == "computer":
            from_computer.put(data) # To be handled by interpreter.computer.run
        elif data["role"] == "user":
            await from_user.put(data)
        else:
            raise("Unknown role:", data)

async def send_messages(websocket: WebSocket):
    while True:
        message = await to_device.get()
        logger.debug(f"Sending to the device: {type(message)} {message}")
        await websocket.send_json(message)

async def listener():
    audio_bytes = bytearray()
    while True:
        while True:
            if not from_user.empty():
                message = await from_user.get()
                break
            elif not from_computer.empty():
                message = from_computer.get()
                break
            await asyncio.sleep(1)

        if type(message) == str:
            message = json.loads(message)

        # Hold the audio in a buffer. If it's ready (we got end flag, stt it)
        if message["type"] == "audio":
            if "content" in message:
                audio_bytes.extend(bytes(ast.literal_eval(message["content"])))
            if "end" in message:
                content = stt_bytes(audio_bytes, message["format"])
                if content == None: # If it was nothing / silence
                    continue
                audio_bytes = bytearray()
                message = {"role": "user", "type": "message", "content": content}
            else:
                continue

        # Ignore flags, we only needed them for audio ^
        if "content" not in message or message["content"] == None:
            continue

        # Custom stop message will halt us
        if message["content"].lower().strip(".,!") == "stop":
            continue

        # Load, append, and save conversation history
        with open(conversation_history_path, 'r') as file:
            messages = json.load(file)
        messages.append(message)
        with open(conversation_history_path, 'w') as file:
            json.dump(messages, file, indent=4)

        accumulated_text = ""
        
        for chunk in interpreter.chat(messages, stream=True, display=False):

            logger.debug("Got chunk:", chunk)

            # Send it to the user
            await to_device.put(chunk)
            # Yield to the event loop, so you actually send it out
            await asyncio.sleep(0.01)
            
            # Speak full sentences out loud
            if chunk["role"] == "assistant" and "content" in chunk:
                accumulated_text += chunk["content"]
                sentences = split_into_sentences(accumulated_text)
                if is_full_sentence(sentences[-1]):
                    for sentence in sentences:
                        await stream_or_play_tts(sentence)
                    accumulated_text = ""
                else:
                    for sentence in sentences[:-1]:
                        await stream_or_play_tts(sentence)
                    accumulated_text = sentences[-1]
            
            # If we have a new message, save our progress and go back to the top
            if not from_user.empty():

                # Check if it's just an end flag. We ignore those.
                temp_message = await from_user.get()
                
                if temp_message == {'role': 'user', 'type': 'message', 'end': True}:
                    # Yup. False alarm.
                    continue
                else:
                    # Whoops! Put that back
                    await from_user.put(temp_message)

                with open(conversation_history_path, 'w') as file:
                    json.dump(interpreter.messages, file, indent=4)

                logger.info("New user message recieved. Breaking.")
                break

            # Also check if there's any new computer messages
            if not from_computer.empty():
                
                with open(conversation_history_path, 'w') as file:
                    json.dump(interpreter.messages, file, indent=4)

                logger.info("New computer message recieved. Breaking.")
                break
        else:
            with open(conversation_history_path, 'w') as file:
                json.dump(interpreter.messages, file, indent=4)
        

async def stream_or_play_tts(sentence):

    if os.getenv('TTS_RUNNER') == "server":
        tts(sentence, play_audio=True)
    else:
        await to_device.put({"role": "assistant", "type": "audio", "format": "audio/mp3", "start": True})
        audio_bytes = tts(sentence, play_audio=False)
        await to_device.put({"role": "assistant", "type": "audio", "format": "audio/mp3", "content": str(audio_bytes)})
        await to_device.put({"role": "assistant", "type": "audio", "format": "audio/mp3", "end": True})

async def setup_ngrok(ngrok_auth_token, parsed_url):
    # Set up Ngrok
    logger.info("Setting up Ngrok")
    ngrok_listener = await ngrok.forward(f"{parsed_url.hostname}:{parsed_url.port}", authtoken=ngrok_auth_token)
    ngrok_parsed_url = urllib.parse.urlparse(ngrok_listener.url())

    # Setup SERVER_URL environment variable for device to use
    connection_url = f"wss://{ngrok_parsed_url.hostname}/"
    logger.info(f"Ngrok established at {ngrok_parsed_url.geturl()}")
    logger.info(f"\033[1mSERVER_CONNECTION_URL should be set to \"{connection_url}\"\033[0m")


from uvicorn import Config, Server

# Run the FastAPI app
if __name__ == "__main__":

    async def main():
        # Start listening
        asyncio.create_task(listener())

        # Start watching the kernel if it's your job to do that
        if os.getenv('CODE_RUNNER') == "server":
            asyncio.create_task(put_kernel_messages_into_queue(from_computer))

        server_url = os.getenv('SERVER_URL')
        if not server_url:
            raise ValueError("The environment variable SERVER_URL is not set. Please set it to proceed.")
        parsed_url = urllib.parse.urlparse(server_url)

        # Set up Ngrok
        ngrok_auth_token = os.getenv('NGROK_AUTHTOKEN')
        if ngrok_auth_token is not None:
            await setup_ngrok(ngrok_auth_token, parsed_url)
            
        logger.info("Starting `server.py`...")

        config = Config(app, host=parsed_url.hostname, port=parsed_url.port, lifespan='on')
        server = Server(config)
        await server.serve()

    asyncio.run(main())